import enum
from datetime import datetime as dt

from gcloud.rest.datastore.lat_lng import LatLng


class CompositeFilterOperator(enum.Enum):
    AND = 'AND'
    UNSPECIFIED = 'OPERATOR_UNSPECIFIED'


class Consistency(enum.Enum):
    EVENTUAL = 'EVENTUAL'
    STRONG = 'STRONG'
    UNSPECIFIED = 'READ_CONSISTENCY_UNSPECIFIED'


class Direction(enum.Enum):
    ASCENDING = 'ASCENDING'
    DESCENDING = 'DESCENDING'
    UNSPECIFIED = 'DIRECTION_UNSPECIFIED'


class Mode(enum.Enum):
    NON_TRANSACTIONAL = 'NON_TRANSACTIONAL'
    TRANSACTIONAL = 'TRANSACTIONAL'
    UNSPECIFIED = 'MODE_UNSPECIFIED'


class MoreResultsType(enum.Enum):
    MORE_RESULTS_AFTER_CURSOR = 'MORE_RESULTS_AFTER_CURSOR'
    MORE_RESULTS_AFTER_LIMIT = 'MORE_RESULTS_AFTER_LIMIT'
    NO_MORE_RESULTS = 'NO_MORE_RESULTS'
    NOT_FINISHED = 'NOT_FINISHED'
    UNSPECIFIED = 'MORE_RESULTS_TYPE_UNSPECIFIED'


class Operation(enum.Enum):
    DELETE = 'delete'
    INSERT = 'insert'
    UPDATE = 'update'
    UPSERT = 'upsert'


class PropertyFilterOperator(enum.Enum):
    EQUAL = 'EQUAL'
    GREATER_THAN = 'GREATER_THAN'
    GREATER_THAN_OR_EQUAL = 'GREATER_THAN_OR_EQUAL'
    HAS_ANCESTOR = 'HAS_ANCESTOR'
    LESS_THAN = 'LESS_THAN'
    LESS_THAN_OR_EQUAL = 'LESS_THAN_OR_EQUAL'
    UNSPECIFIED = 'OPERATOR_UNSPECIFIED'


class ResultType(enum.Enum):
    FULL = 'FULL'
    KEY_ONLY = 'KEY_ONLY'
    PROJECTION = 'PROJECTION'
    UNSPECIFIED = 'RESULT_TYPE_UNSPECIFIED'


class TypeName(enum.Enum):
    ARRAY = 'arrayValue'
    BLOB = 'blobValue'
    BOOLEAN = 'booleanValue'
    DOUBLE = 'doubleValue'
    ENTITY = 'entityValue'
    INTEGER = 'integerValue'
    GEOPOINT = 'geoPointValue'
    KEY = 'keyValue'
    NULL = 'nullValue'
    STRING = 'stringValue'
    TIMESTAMP = 'timestampValue'


TYPES = {
    bool: TypeName.BOOLEAN,
    bytes: TypeName.BLOB,
    dt: TypeName.TIMESTAMP,
    float: TypeName.DOUBLE,
    int: TypeName.INTEGER,
    LatLng: TypeName.GEOPOINT,
    str: TypeName.STRING,
    type(None): TypeName.NULL,
}



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from builtins import super
from builtins import str
import warnings
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from future import standard_library
standard_library.install_aliases()
try:
    from collections.abc import Sequence
except ImportError:
    from collections import Sequence
from typing import Any
from typing import Dict
from typing import List

from gcloud.rest.datastore import value


# https://cloud.google.com/datastore/docs/reference/data/rest/v1/projects/runQuery#ArrayValue
class Array(Sequence):  # type: ignore[type-arg]
    def __init__(self, items                   )        :
        super(Sequence, self).__init__()  # pylint: disable=bad-super-call
        self.items = items

    def __eq__(self, other     )        :
        if not isinstance(other, Array):
            return False
        return self.items == other.items

    def __repr__(self)       :
        return str(self.to_repr())

    def __getitem__(self, index     )       :
        return self.items[index]

    def __len__(self)       :
        return len(self.items)

    @classmethod
    def from_repr(cls, data                )           :
        return cls([value.Value.from_repr(x) for x in data.get('values', [])])

    def to_repr(self)                  :
        return {'values': [x.to_repr() for x in self]}



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from builtins import str
import warnings
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from future import standard_library
standard_library.install_aliases()
from builtins import object
import io
import json
import logging
import os
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

from gcloud.rest.auth import SyncSession  # pylint: disable=no-name-in-module
from gcloud.rest.auth import BUILD_GCLOUD_REST  # pylint: disable=no-name-in-module
from gcloud.rest.auth import Token  # pylint: disable=no-name-in-module
from gcloud.rest.datastore.constants import Consistency
from gcloud.rest.datastore.constants import Mode
from gcloud.rest.datastore.constants import Operation
from gcloud.rest.datastore.datastore_operation import DatastoreOperation
from gcloud.rest.datastore.entity import EntityResult
from gcloud.rest.datastore.key import Key
from gcloud.rest.datastore.mutation import MutationResult
from gcloud.rest.datastore.query import BaseQuery
from gcloud.rest.datastore.query import QueryResultBatch
from gcloud.rest.datastore.value import Value

# Selectively load libraries based on the package
if BUILD_GCLOUD_REST:
    from requests import Session
else:
    from aiohttp import ClientSession as Session   # type: ignore[no-redef]


try:
    API_ROOT = 'http://{}/v1'.format((os.environ["DATASTORE_EMULATOR_HOST"]))
    IS_DEV = True
except KeyError:
    API_ROOT = 'https://datastore.googleapis.com/v1'
    IS_DEV = False

SCOPES = [
    'https://www.googleapis.com/auth/cloud-platform',
    'https://www.googleapis.com/auth/datastore',
]

log = logging.getLogger(__name__)


class Datastore(object):
    datastore_operation_kind = DatastoreOperation
    entity_result_kind = EntityResult
    key_kind = Key
    mutation_result_kind = MutationResult
    query_result_batch_kind = QueryResultBatch
    value_kind = Value

    #_project: Optional[str]

    def __init__(self, project                = None,
                 service_file                                  = None,
                 namespace      = '', session                    = None,
                 token                  = None)        :
        self.namespace = namespace
        self.session = SyncSession(session)

        if IS_DEV:
            self._project = os.environ.get('DATASTORE_PROJECT_ID', 'dev')
            # Tokens are not needed when using dev emulator
            self.token = None
        else:
            self._project = project
            self.token = token or Token(service_file=service_file,
                                        session=self.session.session,
                                        scopes=SCOPES)

    def project(self)       :
        if self._project:
            return self._project

        if IS_DEV or self.token is None:
            raise Exception('project can not be determined in dev mode')

        self._project = self.token.get_project()
        if self._project:
            return self._project

        raise Exception('could not determine project, please set it manually')

    @staticmethod
    def _make_commit_body(mutations                      ,
                          transaction                = None,
                          mode       = Mode.TRANSACTIONAL)                  :
        if not mutations:
            raise Exception('at least one mutation record is required')

        if transaction is None and mode != Mode.NON_TRANSACTIONAL:
            raise Exception('a transaction ID must be provided when mode is '
                            'transactional')

        data = {
            'mode': mode.value,
            'mutations': mutations,
        }
        if transaction is not None:
            data['transaction'] = transaction
        return data

    def headers(self)                  :
        if IS_DEV or self.token is None:
            return {}

        token = self.token.get()
        return {
            'Authorization': 'Bearer {}'.format((token)),
        }

    # TODO: support mutations w version specifiers, return new version (commit)
    @classmethod
    def make_mutation(
            cls, operation           , key     ,
            properties                           = None)                  :
        if operation == Operation.DELETE:
            return {operation.value: key.to_repr()}

        return {
            operation.value: {
                'key': key.to_repr(),
                'properties': {k: cls.value_kind(v).to_repr()
                               for k, v in (properties or {}).items()},
            }
        }

    # https://cloud.google.com/datastore/docs/reference/data/rest/v1/projects/allocateIds
    def allocateIds(self, keys           ,
                          session                    = None,
                          timeout      = 10)             :
        project = self.project()
        url = '{}/projects/{}:allocateIds'.format((API_ROOT), (project))

        payload = json.dumps({
            'keys': [k.to_repr() for k in keys],
        }).encode('utf-8')

        headers = self.headers()
        headers.update({
            'Content-Length': str(len(payload)),
            'Content-Type': 'application/json',
        })

        s = SyncSession(session) if session else self.session
        resp = s.post(url, data=payload, headers=headers,
                            timeout=timeout)
        data = resp.json()

        return [self.key_kind.from_repr(k) for k in data['keys']]

    # https://cloud.google.com/datastore/docs/reference/data/rest/v1/projects/beginTransaction
    # TODO: support readwrite vs readonly transaction types
    def beginTransaction(self, session                    = None,
                               timeout      = 10)       :
        project = self.project()
        url = '{}/projects/{}:beginTransaction'.format((API_ROOT), (project))
        headers = self.headers()
        headers.update({
            'Content-Length': '0',
            'Content-Type': 'application/json',
        })

        s = SyncSession(session) if session else self.session
        resp = s.post(url, headers=headers, timeout=timeout)
        data = resp.json()

        transaction      = data['transaction']
        return transaction

    # https://cloud.google.com/datastore/docs/reference/data/rest/v1/projects/commit
    def commit(self, mutations                      ,
                     transaction                = None,
                     mode       = Mode.TRANSACTIONAL,
                     session                    = None,
                     timeout      = 10)                  :
        project = self.project()
        url = '{}/projects/{}:commit'.format((API_ROOT), (project))

        body = self._make_commit_body(mutations, transaction=transaction,
                                      mode=mode)
        payload = json.dumps(body).encode('utf-8')

        headers = self.headers()
        headers.update({
            'Content-Length': str(len(payload)),
            'Content-Type': 'application/json',
        })

        s = SyncSession(session) if session else self.session
        resp = s.post(url, data=payload, headers=headers,
                            timeout=timeout)
        data                 = resp.json()

        return {
            'mutationResults': [self.mutation_result_kind.from_repr(r)
                                for r in data.get('mutationResults', [])],
            'indexUpdates': data.get('indexUpdates', 0),
        }

    # https://cloud.google.com/datastore/docs/reference/admin/rest/v1/projects/export
    def export(self, output_bucket_prefix     ,
                     kinds                      = None,
                     namespaces                      = None,
                     labels                           = None,
                     session                    = None,
                     timeout      = 10)                      :
        project = self.project()
        url = '{}/projects/{}:export'.format((API_ROOT), (project))

        payload = json.dumps({
            'entityFilter': {
                'kinds': kinds or [],
                'namespaceIds': namespaces or [],
            },
            'labels': labels or {},
            'outputUrlPrefix': 'gs://{}'.format((output_bucket_prefix)),
        }).encode('utf-8')

        headers = self.headers()
        headers.update({
            'Content-Length': str(len(payload)),
            'Content-Type': 'application/json',
        })

        s = SyncSession(session) if session else self.session
        resp = s.post(url, data=payload, headers=headers,
                            timeout=timeout)
        data                 = resp.json()

        return self.datastore_operation_kind.from_repr(data)

    # https://cloud.google.com/datastore/docs/reference/data/rest/v1/projects.operations/get
    def get_datastore_operation(self, name     ,
                                      session                    = None,
                                      timeout      = 10)                      :
        url = '{}/{}'.format((API_ROOT), (name))

        headers = self.headers()
        headers.update({
            'Content-Type': 'application/json',
        })

        s = SyncSession(session) if session else self.session
        resp = s.get(url, headers=headers, timeout=timeout)
        data                 = resp.json()

        return self.datastore_operation_kind.from_repr(data)

    # https://cloud.google.com/datastore/docs/reference/data/rest/v1/projects/lookup
    def lookup(
            self, keys           , transaction                = None,
            consistency              = Consistency.STRONG,
            session                    = None, timeout      = 10
    )                                             :
        project = self.project()
        url = '{}/projects/{}:lookup'.format((API_ROOT), (project))

        if transaction:
            options = {'transaction': transaction}
        else:
            options = {'readConsistency': consistency.value}
        payload = json.dumps({
            'keys': [k.to_repr() for k in keys],
            'readOptions': options,
        }).encode('utf-8')

        headers = self.headers()
        headers.update({
            'Content-Length': str(len(payload)),
            'Content-Type': 'application/json',
        })

        s = SyncSession(session) if session else self.session
        resp = s.post(url, data=payload, headers=headers,
                            timeout=timeout)

        data                       = resp.json()

        return {
            'found': [self.entity_result_kind.from_repr(e)
                      for e in data.get('found', [])],
            'missing': [self.entity_result_kind.from_repr(e)
                        for e in data.get('missing', [])],
            'deferred': [self.key_kind.from_repr(k)
                         for k in data.get('deferred', [])],
        }

    # https://cloud.google.com/datastore/docs/reference/data/rest/v1/projects/reserveIds
    def reserveIds(self, keys           , database_id      = '',
                         session                    = None,
                         timeout      = 10)        :
        project = self.project()
        url = '{}/projects/{}:reserveIds'.format((API_ROOT), (project))

        payload = json.dumps({
            'databaseId': database_id,
            'keys': [k.to_repr() for k in keys],
        }).encode('utf-8')

        headers = self.headers()
        headers.update({
            'Content-Length': str(len(payload)),
            'Content-Type': 'application/json',
        })

        s = SyncSession(session) if session else self.session
        s.post(url, data=payload, headers=headers, timeout=timeout)

    # https://cloud.google.com/datastore/docs/reference/data/rest/v1/projects/rollback
    def rollback(self, transaction     ,
                       session                    = None,
                       timeout      = 10)        :
        project = self.project()
        url = '{}/projects/{}:rollback'.format((API_ROOT), (project))

        payload = json.dumps({
            'transaction': transaction,
        }).encode('utf-8')

        headers = self.headers()
        headers.update({
            'Content-Length': str(len(payload)),
            'Content-Type': 'application/json',
        })

        s = SyncSession(session) if session else self.session
        s.post(url, data=payload, headers=headers, timeout=timeout)

    # https://cloud.google.com/datastore/docs/reference/data/rest/v1/projects/runQuery
    def runQuery(self, query           ,
                       transaction                = None,
                       consistency              = Consistency.EVENTUAL,
                       session                    = None,
                       timeout      = 10)                    :
        project = self.project()
        url = '{}/projects/{}:runQuery'.format((API_ROOT), (project))

        if transaction:
            options = {'transaction': transaction}
        else:
            options = {'readConsistency': consistency.value}
        payload = json.dumps({
            'partitionId': {
                'projectId': project,
                'namespaceId': self.namespace,
            },
            query.json_key: query.to_repr(),
            'readOptions': options,
        }).encode('utf-8')

        headers = self.headers()
        headers.update({
            'Content-Length': str(len(payload)),
            'Content-Type': 'application/json',
        })

        s = SyncSession(session) if session else self.session
        resp = s.post(url, data=payload, headers=headers,
                            timeout=timeout)

        data                 = resp.json()
        return self.query_result_batch_kind.from_repr(data['batch'])

    def delete(self, key     ,
                     session                    = None)                  :
        return self.operate(Operation.DELETE, key, session=session)

    def insert(self, key     , properties                ,
                     session                    = None)                  :
        return self.operate(Operation.INSERT, key, properties,
                                  session=session)

    def update(self, key     , properties                ,
                     session                    = None)                  :
        return self.operate(Operation.UPDATE, key, properties,
                                  session=session)

    def upsert(self, key     , properties                ,
                     session                    = None)                  :
        return self.operate(Operation.UPSERT, key, properties,
                                  session=session)

    # TODO: accept Entity rather than key/properties?
    def operate(self, operation           , key     ,
                      properties                           = None,
                      session                    = None)                  :
        transaction = self.beginTransaction(session=session)
        mutation = self.make_mutation(operation, key, properties=properties)
        return self.commit([mutation], transaction=transaction,
                                 session=session)

    def close(self)        :
        self.session.close()

    def __enter__(self)               :
        return self

    def __exit__(self, *args     )        :
        self.close()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from builtins import str
from builtins import range
import warnings
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from future import standard_library
standard_library.install_aliases()
import uuid

import pytest
from gcloud.rest.auth import BUILD_GCLOUD_REST  # pylint: disable=no-name-in-module
from gcloud.rest.datastore import Datastore
from gcloud.rest.datastore import Filter
from gcloud.rest.datastore import GQLQuery
from gcloud.rest.datastore import Key
from gcloud.rest.datastore import Operation
from gcloud.rest.datastore import PathElement
from gcloud.rest.datastore import Projection
from gcloud.rest.datastore import PropertyFilter
from gcloud.rest.datastore import PropertyFilterOperator
from gcloud.rest.datastore import Query
from gcloud.rest.datastore import Value
from gcloud.rest.storage import Storage  # pylint: disable=no-name-in-module

# Selectively load libraries based on the package
if BUILD_GCLOUD_REST:
    from requests import Session
    from time import sleep
else:
    from aiohttp import ClientSession as Session
    from asyncio import sleep


#@pytest.mark.asyncio  # type: ignore
def test_item_lifecycle(creds     , kind     , project     )        :
    key = Key(project, [PathElement(kind)])

    with Session() as s:
        ds = Datastore(project=project, service_file=creds, session=s)

        allocatedKeys = ds.allocateIds([key], session=s)
        assert len(allocatedKeys) == 1
        key.path[-1].id = allocatedKeys[0].path[-1].id
        assert key == allocatedKeys[0]

        ds.reserveIds(allocatedKeys, session=s)

        props_insert = {'is_this_bad_data': True}
        ds.insert(allocatedKeys[0], props_insert, session=s)
        actual = ds.lookup([allocatedKeys[0]], session=s)
        assert actual['found'][0].entity.properties == props_insert

        props_update = {'animal': 'aardvark', 'overwrote_bad_data': True}
        ds.update(allocatedKeys[0], props_update, session=s)
        actual = ds.lookup([allocatedKeys[0]], session=s)
        assert actual['found'][0].entity.properties == props_update

        props_upsert = {'meaning_of_life': 42}
        ds.upsert(allocatedKeys[0], props_upsert, session=s)
        actual = ds.lookup([allocatedKeys[0]], session=s)
        assert actual['found'][0].entity.properties == props_upsert

        ds.delete(allocatedKeys[0], session=s)
        actual = ds.lookup([allocatedKeys[0]], session=s)
        assert len(actual['missing']) == 1


#@pytest.mark.asyncio  # type: ignore
def test_mutation_result(creds     , kind     , project     )        :
    key = Key(project, [PathElement(kind)])

    with Session() as s:
        ds = Datastore(project=project, service_file=creds, session=s)

        insert_result = ds.insert(key, {'value': 12})
        assert len(insert_result['mutationResults']) == 1
        saved_key = insert_result['mutationResults'][0].key
        assert saved_key is not None

        update_result = ds.update(saved_key, {'value': 83})
        assert len(update_result['mutationResults']) == 1
        assert update_result['mutationResults'][0].key is None

        delete_result = ds.delete(saved_key)
        assert len(delete_result['mutationResults']) == 1
        assert delete_result['mutationResults'][0].key is None


#@pytest.mark.asyncio  # type: ignore
def test_transaction(creds     , kind     , project     )        :
    key = Key(project, [PathElement(kind, name='test_record_{}'.format((uuid.uuid4())))])

    with Session() as s:
        ds = Datastore(project=project, service_file=creds, session=s)

        transaction = ds.beginTransaction(session=s)
        actual = ds.lookup([key], transaction=transaction, session=s)
        assert len(actual['missing']) == 1

        mutations = [
            ds.make_mutation(Operation.INSERT, key,
                             properties={'animal': 'three-toed sloth'}),
            ds.make_mutation(Operation.UPDATE, key,
                             properties={'animal': 'aardvark'}),
        ]
        ds.commit(mutations, transaction=transaction, session=s)

        actual = ds.lookup([key], session=s)
        assert actual['found'][0].entity.properties == {'animal': 'aardvark'}


#@pytest.mark.asyncio  # type: ignore
def test_rollback(creds     , project     )        :
    with Session() as s:
        ds = Datastore(project=project, service_file=creds, session=s)

        transaction = ds.beginTransaction(session=s)
        ds.rollback(transaction, session=s)


#@pytest.mark.asyncio  # type: ignore
def test_query_with_key_projection(creds     , kind     ,
                                         project     )        :
    with Session() as s:
        ds = Datastore(project=project, service_file=creds, session=s)
        # setup test data
        ds.insert(Key(project, [PathElement(kind)]), {'value': 30}, s)
        property_filter = PropertyFilter(
            prop='value', operator=PropertyFilterOperator.EQUAL,
            value=Value(30))
        projection = [Projection.from_repr({'property': {'name': '__key__'}})]

        query = Query(kind=kind, query_filter=Filter(property_filter), limit=1,
                      projection=projection)
        result = ds.runQuery(query, session=s)
        assert result.entity_results[0].entity.properties == {}
        assert result.entity_result_type.value == 'KEY_ONLY'
        # clean up test data
        ds.delete(result.entity_results[0].entity.key, s)


#@pytest.mark.asyncio  # type: ignore
def test_query_with_value_projection(creds     , kind     ,
                                           project     )        :
    with Session() as s:
        ds = Datastore(project=project, service_file=creds, session=s)
        # setup test data
        ds.insert(Key(project, [PathElement(kind)]), {'value': 30}, s)
        projection = [Projection.from_repr({'property': {'name': 'value'}})]

        query = Query(kind=kind, limit=1,
                      projection=projection)
        result = ds.runQuery(query, session=s)
        assert result.entity_result_type.value == 'PROJECTION'
        # clean up test data
        ds.delete(result.entity_results[0].entity.key, s)


#@pytest.mark.asyncio  # type: ignore
def test_query_with_distinct_on(creds     , kind     ,
                                      project     )        :
    keys1 = [Key(project, [PathElement(kind)]) for i in range(3)]
    keys2 = [Key(project, [PathElement(kind)]) for i in range(3)]
    with Session() as s:
        ds = Datastore(project=project, service_file=creds, session=s)

        # setup test data
        allocatedKeys1 = ds.allocateIds(keys1, session=s)
        allocatedKeys2 = ds.allocateIds(keys2, session=s)
        for key1 in allocatedKeys1:
            ds.insert(key1, {'dist_value': 11}, s)
        for key2 in allocatedKeys2:
            ds.insert(key2, {'dist_value': 22}, s)
        query = Query(kind=kind, limit=10, distinct_on=['dist_value'])
        result = ds.runQuery(query, session=s)
        assert len(result.entity_results) == 2
        # clean up test data
        for key1 in allocatedKeys1:
            ds.delete(key1, s)
        for key2 in allocatedKeys2:
            ds.delete(key2, s)


#@pytest.mark.asyncio  # type: ignore
@pytest.mark.xfail(strict=False)  # type: ignore
def test_query(creds     , kind     , project     )        :
    with Session() as s:
        ds = Datastore(project=project, service_file=creds, session=s)

        property_filter = PropertyFilter(
            prop='value', operator=PropertyFilterOperator.EQUAL,
            value=Value(42))
        query = Query(kind=kind, query_filter=Filter(property_filter))

        before = ds.runQuery(query, session=s)
        num_results = len(before.entity_results)

        transaction = ds.beginTransaction(session=s)
        mutations = [
            ds.make_mutation(Operation.INSERT,
                             Key(project, [PathElement(kind)]),
                             properties={'value': 42}),
            ds.make_mutation(Operation.INSERT,
                             Key(project, [PathElement(kind)]),
                             properties={'value': 42}),
        ]
        ds.commit(mutations, transaction=transaction, session=s)

        after = ds.runQuery(query, session=s)
        assert len(after.entity_results) == num_results + 2


#@pytest.mark.asyncio  # type: ignore
@pytest.mark.xfail(strict=False)  # type: ignore
def test_gql_query(creds     , kind     , project     )        :
    with Session() as s:
        ds = Datastore(project=project, service_file=creds, session=s)

        query = GQLQuery('SELECT * FROM {} WHERE value = @value'.format((kind)),
                         named_bindings={'value': 42})

        before = ds.runQuery(query, session=s)
        num_results = len(before.entity_results)

        transaction = ds.beginTransaction(session=s)
        mutations = [
            ds.make_mutation(Operation.INSERT,
                             Key(project, [PathElement(kind)]),
                             properties={'value': 42}),
            ds.make_mutation(Operation.INSERT,
                             Key(project, [PathElement(kind)]),
                             properties={'value': 42}),
            ds.make_mutation(Operation.INSERT,
                             Key(project, [PathElement(kind)]),
                             properties={'value': 42}),
        ]
        ds.commit(mutations, transaction=transaction, session=s)

        after = ds.runQuery(query, session=s)
        assert len(after.entity_results) == num_results + 3


#@pytest.mark.asyncio  # type: ignore
def test_datastore_export(creds     , project     ,
                                export_bucket_name     ):
    # N.B. when modifying this test, please also see `test_table_load_copy` in
    # `gcloud-rest-bigquery`.
    kind = 'PublicTestDatastoreExportModel'

    rand_uuid = str(uuid.uuid4())

    with Session() as s:
        ds = Datastore(project=project, service_file=creds, session=s)

        ds.insert(Key(project, [PathElement(kind)]),
                        properties={'rand_str': rand_uuid})

        operation = ds.export(export_bucket_name, kinds=[kind])

        count = 0
        while (count < 10
               and operation
               and operation.metadata['common']['state'] == 'PROCESSING'):
            sleep(10)
            operation = ds.get_datastore_operation(operation.name)
            count += 1

        assert operation.metadata['common']['state'] == 'SUCCESSFUL'

        prefix_len = len('gs://{}/'.format((export_bucket_name)))
        export_path = operation.metadata['outputUrlPrefix'][prefix_len:]

        storage = Storage(service_file=creds, session=s)
        files = storage.list_objects(export_bucket_name,
                                           params={'prefix': export_path})
        for file in files['items']:
            storage.delete(export_bucket_name, file['name'])



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from builtins import str
import warnings
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from future import standard_library
standard_library.install_aliases()
from builtins import object
from typing import Any
from typing import Dict
from typing import List
from typing import Optional


class PathElement(object):
    def __init__(self, kind     , **_3to2kwargs)        :
        if 'name' in _3to2kwargs: name = _3to2kwargs['name']; del _3to2kwargs['name']
        else: name =  None
        if 'id_' in _3to2kwargs: id_ = _3to2kwargs['id_']; del _3to2kwargs['id_']
        else: id_ =  None
        self.kind = kind

        self.id = id_
        self.name = name
        if self.id and self.name:
            raise Exception('invalid PathElement contains both ID and name')

    def __eq__(self, other     )        :
        if not isinstance(other, PathElement):
            return False

        return bool(self.kind == other.kind and self.id == other.id
                    and self.name == other.name)

    def __repr__(self)       :
        return str(self.to_repr())

    @classmethod
    def from_repr(cls, data                )                 :
        kind      = data['kind']
        id_                = data.get('id')
        name                = data.get('name')
        return cls(kind, id_=id_, name=name)

    def to_repr(self)                  :
        data                 = {'kind': self.kind}
        if self.id:
            data['id'] = self.id
        elif self.name:
            data['name'] = self.name

        return data


class Key(object):
    path_element_kind = PathElement

    def __init__(self, project     , path                   ,
                 namespace      = '')        :
        self.project = project
        self.namespace = namespace
        self.path = path

    def __eq__(self, other     )        :
        if not isinstance(other, Key):
            return False

        return bool(self.project == other.project
                    and self.namespace == other.namespace
                    and self.path == other.path)

    def __repr__(self)       :
        return str(self.to_repr())

    @classmethod
    def from_repr(cls, data                )         :
        return cls(data['partitionId']['projectId'],
                   path=[cls.path_element_kind.from_repr(p)
                         for p in data['path']],
                   namespace=data['partitionId'].get('namespaceId', ''))

    def to_repr(self)                  :
        return {
            'partitionId': {
                'projectId': self.project,
                'namespaceId': self.namespace,
            },
            'path': [p.to_repr() for p in self.path],
        }



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
import warnings
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from future import standard_library
standard_library.install_aliases()
try:
    import pkg_resources
    pkg_resources.declare_namespace(__name__)
except ImportError:
    import pkgutil
    __path__ = pkgutil.extend_path(
        __path__, __name__)  # type: ignore[has-type]



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from builtins import str
import warnings
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from future import standard_library
standard_library.install_aliases()
from builtins import object
from typing import Any
from typing import Dict
from typing import Optional

from gcloud.rest.datastore.key import Key


# https://cloud.google.com/datastore/docs/reference/data/rest/v1/projects/commit#Mutation
class Mutation(object):
    # TODO: Use this Mutation class instead of datastore.make_mutation
    pass


# https://cloud.google.com/datastore/docs/reference/data/rest/v1/projects/commit#MutationResult
class MutationResult(object):
    key_kind = Key

    def __init__(self, key               , version     ,
                 conflict_detected      )        :
        self.key = key
        self.version = version
        self.conflict_detected = conflict_detected

    def __eq__(self, other     )        :
        if not isinstance(other, MutationResult):
            return False

        return bool(self.key == other.key and self.version == other.version)

    def __repr__(self)       :
        return str(self.to_repr())

    @classmethod
    def from_repr(cls, data                )                    :
        if 'key' in data:
            key                = cls.key_kind.from_repr(data['key'])
        else:
            key = None
        version      = data['version']
        conflict_detected       = data.get('conflictDetected', False)
        return cls(key, version, conflict_detected)

    def to_repr(self)                  :
        data = {
            'version': self.version,
            'conflictDetected': self.conflict_detected
        }
        if self.key:
            data['key'] = self.key.to_repr()
        return data



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from builtins import str
import warnings
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from future import standard_library
standard_library.install_aliases()
from builtins import object
from typing import Any
from typing import Dict
from typing import Optional

from gcloud.rest.datastore.key import Key
from gcloud.rest.datastore.value import Value


class Entity(object):
    key_kind = Key
    value_kind = Value

    def __init__(
            self, key               ,
            properties                                      = None)        :
        self.key = key
        self.properties = {k: self.value_kind.from_repr(v).value
                           for k, v in (properties or {}).items()}

    def __eq__(self, other     )        :
        if not isinstance(other, Entity):
            return False

        return bool(self.key == other.key
                    and self.properties == other.properties)

    def __repr__(self)       :
        return str(self.to_repr())

    @classmethod
    def from_repr(cls, data                )            :
        # https://cloud.google.com/datastore/docs/reference/data/rest/v1/Entity
        # "for example, an entity in Value.entity_value may have no key"
        if 'key' in data:
            key                = cls.key_kind.from_repr(data['key'])
        else:
            key = None
        return cls(key, data.get('properties'))

    def to_repr(self)                  :
        return {
            'key': self.key.to_repr() if self.key else None,
            'properties': {k: self.value_kind(v).to_repr()
                           for k, v in self.properties.items()},
        }


class EntityResult(object):
    entity_kind = Entity

    def __init__(self, entity        , version      = '',
                 cursor      = '')        :
        self.entity = entity
        self.version = version
        self.cursor = cursor

    def __eq__(self, other     )        :
        if not isinstance(other, EntityResult):
            return False

        return bool(self.entity == other.entity
                    and self.version == other.version
                    and self.cursor == self.cursor)

    def __repr__(self)       :
        return str(self.to_repr())

    @classmethod
    def from_repr(cls, data                )                  :
        return cls(cls.entity_kind.from_repr(data['entity']),
                   data.get('version', ''),
                   data.get('cursor', ''))

    def to_repr(self)                  :
        data                 = {
            'entity': self.entity.to_repr(),
        }
        if self.version:
            data['version'] = self.version
        if self.cursor:
            data['cursor'] = self.cursor

        return data



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from builtins import str
from builtins import int
import warnings
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from future import standard_library
standard_library.install_aliases()
from builtins import object
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

from gcloud.rest.datastore.constants import MoreResultsType
from gcloud.rest.datastore.constants import ResultType
from gcloud.rest.datastore.entity import EntityResult
from gcloud.rest.datastore.filter import Filter
from gcloud.rest.datastore.projection import Projection
from gcloud.rest.datastore.property_order import PropertyOrder
from gcloud.rest.datastore.value import Value


class BaseQuery(object):
    #json_key: str
    value_kind = Value

    def __repr__(self)       :
        return str(self.to_repr())

    @classmethod
    def from_repr(cls, data                )               :
        raise NotImplementedError

    def to_repr(self)                  :
        raise NotImplementedError


# https://cloud.google.com/datastore/docs/reference/data/rest/v1/projects/runQuery#Query
class Query(BaseQuery):
    # pylint: disable=too-many-instance-attributes
    json_key = 'query'

    def __init__(self, kind      = '', query_filter                   = None,
                 order                                = None,
                 start_cursor      = '', end_cursor      = '',
                 offset                = None, limit                = None,
                 projection                             = None,
                 distinct_on                      = None)        :
        self.kind = kind
        self.query_filter = query_filter
        self.orders = order or []
        self.start_cursor = start_cursor
        self.end_cursor = end_cursor
        self.offset = offset
        self.limit = limit
        self.projection = projection or []
        self.distinct_on = distinct_on or []

    def __eq__(self, other     )        :
        if not isinstance(other, Query):
            return False

        return bool(
            self.kind == other.kind
            and self.query_filter == other.query_filter)

    @classmethod
    def from_repr(cls, data                )           :
        kind = data['kind'] or ''  # Kind is required
        orders = [PropertyOrder.from_repr(o) for o in data.get('order', [])]
        start_cursor = data.get('startCursor') or ''
        end_cursor = data.get('endCursor') or ''
        offset = int(data['offset']) if 'offset' in data else None
        limit = int(data['limit']) if 'limit' in data else None
        projection = [Projection.from_repr(p)
                      for p in data.get('projection', [])]
        distinct_on = [d['name'] for d in data.get('distinct_on', [])]

        filter_ = data.get('filter')
        query_filter = Filter.from_repr(filter_) if filter_ else None

        return cls(kind=kind, query_filter=query_filter, order=orders,
                   start_cursor=start_cursor, end_cursor=end_cursor,
                   offset=offset, limit=limit,
                   projection=projection, distinct_on=distinct_on)

    def to_repr(self)                  :
        data                 = {
            'kind': [{'name': self.kind}] if self.kind else [],
        }
        if self.query_filter:
            data['filter'] = self.query_filter.to_repr()
        if self.orders:
            data['order'] = [o.to_repr() for o in self.orders]
        if self.start_cursor:
            data['startCursor'] = self.start_cursor
        if self.end_cursor:
            data['endCursor'] = self.end_cursor
        if self.offset is not None:
            data['offset'] = self.offset
        if self.limit is not None:
            data['limit'] = self.limit
        if self.projection:
            data['projection'] = [p.to_repr() for p in self.projection]
        if self.distinct_on:
            data['distinctOn'] = [{'name': d} for d in self.distinct_on]
        return data


# https://cloud.google.com/datastore/docs/reference/data/rest/v1/projects/runQuery#gqlquery
class GQLQuery(BaseQuery):
    json_key = 'gqlQuery'

    def __init__(self, query_string     , allow_literals       = True,
                 named_bindings                           = None,
                 positional_bindings                      = None)        :
        self.query_string = query_string
        self.allow_literals = allow_literals
        self.named_bindings = named_bindings or {}
        self.positional_bindings = positional_bindings or []

    def __eq__(self, other     )        :
        if not isinstance(other, GQLQuery):
            return False

        return bool(
            self.query_string == other.query_string
            and self.allow_literals == other.allow_literals
            and self.named_bindings == other.named_bindings
            and self.positional_bindings == other.positional_bindings)

    @classmethod
    def from_repr(cls, data                )              :
        allow_literals = data['allowLiterals']
        query_string = data['queryString']
        named_bindings = {k: cls.value_kind.from_repr(v['value'].value)
                          for k, v in data.get('namedBindings', {}).items()}
        positional_bindings = [cls.value_kind.from_repr(v['value'].value)
                               for v in data.get('positionalBindings', [])]
        return cls(query_string, allow_literals=allow_literals,
                   named_bindings=named_bindings,
                   positional_bindings=positional_bindings)

    def to_repr(self)                  :
        return {
            'allowLiterals': self.allow_literals,
            'queryString': self.query_string,
            'namedBindings': {k: {'value': self.value_kind(v).to_repr()}
                              for k, v in self.named_bindings.items()},
            'positionalBindings': [{'value': self.value_kind(v).to_repr()}
                                   for v in self.positional_bindings],
        }


class QueryResultBatch(object):
    entity_result_kind = EntityResult

    def __init__(self, end_cursor     ,
                 entity_result_type             = ResultType.UNSPECIFIED,
                 entity_results                               = None,
                 more_results                  = MoreResultsType.UNSPECIFIED,
                 skipped_cursor      = '', skipped_results      = 0,
                 snapshot_version      = '')        :
        self.end_cursor = end_cursor

        self.entity_result_type = entity_result_type
        self.entity_results = entity_results or []
        self.more_results = more_results
        self.skipped_cursor = skipped_cursor
        self.skipped_results = skipped_results
        self.snapshot_version = snapshot_version

    def __eq__(self, other     )        :
        if not isinstance(other, QueryResultBatch):
            return False

        return bool(self.end_cursor == other.end_cursor
                    and self.entity_result_type == other.entity_result_type
                    and self.entity_results == other.entity_results
                    and self.more_results == other.more_results
                    and self.skipped_cursor == other.skipped_cursor
                    and self.skipped_results == other.skipped_results
                    and self.snapshot_version == other.snapshot_version)

    def __repr__(self)       :
        return str(self.to_repr())

    @classmethod
    def from_repr(cls, data                )                      :
        end_cursor = data['endCursor']
        entity_result_type = ResultType(data['entityResultType'])
        entity_results = [cls.entity_result_kind.from_repr(er)
                          for er in data.get('entityResults', [])]
        more_results = MoreResultsType(data['moreResults'])
        skipped_cursor = data.get('skippedCursor', '')
        skipped_results = data.get('skippedResults', 0)
        snapshot_version = data.get('snapshotVersion', '')
        return cls(end_cursor, entity_result_type=entity_result_type,
                   entity_results=entity_results, more_results=more_results,
                   skipped_cursor=skipped_cursor,
                   skipped_results=skipped_results,
                   snapshot_version=snapshot_version)

    def to_repr(self)                  :
        data = {
            'endCursor': self.end_cursor,
            'entityResults': [er.to_repr() for er in self.entity_results],
            'entityResultType': self.entity_result_type.value,
            'moreResults': self.more_results.value,
            'skippedResults': self.skipped_results,
        }
        if self.skipped_cursor:
            data['skippedCursor'] = self.skipped_cursor
        if self.snapshot_version:
            data['snapshotVersion'] = self.snapshot_version

        return data



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
import warnings
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from future import standard_library
standard_library.install_aliases()
from pkg_resources import get_distribution
__version__ = get_distribution('gcloud-rest-datastore').version

from gcloud.rest.datastore.constants import CompositeFilterOperator
from gcloud.rest.datastore.constants import Consistency
from gcloud.rest.datastore.constants import Direction
from gcloud.rest.datastore.constants import Mode
from gcloud.rest.datastore.constants import MoreResultsType
from gcloud.rest.datastore.constants import Operation
from gcloud.rest.datastore.constants import PropertyFilterOperator
from gcloud.rest.datastore.constants import ResultType
from gcloud.rest.datastore.datastore import Datastore
from gcloud.rest.datastore.datastore import SCOPES
from gcloud.rest.datastore.datastore_operation import DatastoreOperation
from gcloud.rest.datastore.entity import Entity
from gcloud.rest.datastore.entity import EntityResult
from gcloud.rest.datastore.filter import CompositeFilter
from gcloud.rest.datastore.filter import Filter
from gcloud.rest.datastore.filter import PropertyFilter
from gcloud.rest.datastore.key import Key
from gcloud.rest.datastore.key import PathElement
from gcloud.rest.datastore.lat_lng import LatLng
from gcloud.rest.datastore.mutation import MutationResult
from gcloud.rest.datastore.projection import Projection
from gcloud.rest.datastore.property_order import PropertyOrder
from gcloud.rest.datastore.query import GQLQuery
from gcloud.rest.datastore.query import Query
from gcloud.rest.datastore.query import QueryResultBatch
from gcloud.rest.datastore.value import Value


__all__ = ['__version__', 'CompositeFilter', 'CompositeFilterOperator',
           'Consistency', 'Datastore', 'DatastoreOperation', 'Direction',
           'Entity', 'EntityResult', 'Filter', 'GQLQuery', 'Key', 'LatLng',
           'Mode', 'MoreResultsType', 'MutationResult', 'Operation',
           'PathElement', 'Projection', 'PropertyFilter',
           'PropertyFilterOperator', 'PropertyOrder', 'Query',
           'QueryResultBatch', 'ResultType', 'SCOPES', 'Value']



"""Make sure all value types are serialized/deserialized correctly"""
import pytest
from gcloud.rest.auth import BUILD_GCLOUD_REST  # pylint:disable=no-name-in-module
from gcloud.rest.datastore import Datastore
from gcloud.rest.datastore import Key
from gcloud.rest.datastore import LatLng
from gcloud.rest.datastore import PathElement

# Selectively load libraries based on the package
if BUILD_GCLOUD_REST:
    from requests import Session
else:
    from aiohttp import ClientSession as Session


#@pytest.mark.asyncio  # type: ignore
def test_geo_point_value(creds     , kind     , project     )        :
    key = Key(project, [PathElement(kind)])

    with Session() as s:
        ds = Datastore(project=project, service_file=creds, session=s)

        allocatedKeys = ds.allocateIds([key], session=s)
        ds.reserveIds(allocatedKeys, session=s)

        props_insert = {'location': LatLng(49.2827, 123.1207)}
        ds.insert(allocatedKeys[0], props_insert, session=s)
        actual = ds.lookup([allocatedKeys[0]], session=s)
        assert actual['found'][0].entity.properties == props_insert



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from builtins import str
import warnings
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from future import standard_library
standard_library.install_aliases()
from builtins import object
from typing import Any
from typing import Dict
from typing import Optional


class DatastoreOperation(object):
    def __init__(self, name     , done      ,
                 metadata                           = None,
                 error                           = None,
                 response                           = None)        :
        self.name = name
        self.done = done

        self.metadata = metadata
        self.error = error
        self.response = response

    def __repr__(self)       :
        return str(self.to_repr())

    @classmethod
    def from_repr(cls, data                )                        :
        return cls(data['name'], data.get('done', False), data.get('metadata'),
                   data.get('error'), data.get('response'))

    def to_repr(self)                  :
        return {
            'done': self.done,
            'error': self.error,
            'metadata': self.metadata,
            'name': self.name,
            'response': self.response,
        }



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from builtins import str
import warnings
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from future import standard_library
standard_library.install_aliases()
from builtins import object
from typing import Any
from typing import Dict
from typing import List

import pytest
from gcloud.rest.datastore import CompositeFilter
from gcloud.rest.datastore import CompositeFilterOperator
from gcloud.rest.datastore import Filter
from gcloud.rest.datastore import PropertyFilter
from gcloud.rest.datastore import PropertyFilterOperator
from gcloud.rest.datastore import Value


class TestFilter(object):
    @staticmethod
    def test_property_filter_from_repr(property_filters):
        original_filter = property_filters[0]
        data = {
            'property': {
                'name': original_filter.prop
            },
            'op': original_filter.operator,
            'value': original_filter.value.to_repr()
        }

        output_filter = PropertyFilter.from_repr(data)

        assert output_filter == original_filter

    def test_property_filter_to_repr(self, property_filters):
        property_filter = property_filters[0]
        query_filter = Filter(inner_filter=property_filter)

        r = query_filter.to_repr()

        self._assert_is_correct_prop_dict_for_property_filter(
            r['propertyFilter'], property_filter)

    @staticmethod
    def test_composite_filter_from_repr(property_filters):
        # pylint: disable=line-too-long
        original_filter = CompositeFilter(
            operator=CompositeFilterOperator.AND,
            filters=[
                Filter(property_filters[0]),
                Filter(property_filters[1]),
            ])
        data = {
            'op': original_filter.operator,
            'filters': [
                {
                    'propertyFilter': {
                        'property': {
                            'name': original_filter.filters[0].inner_filter.prop,
                        },
                        'op': original_filter.filters[0].inner_filter.operator,
                        'value': property_filters[0].value.to_repr(),
                    },
                },
                {
                    'propertyFilter': {
                        'property': {
                            'name': original_filter.filters[1].inner_filter.prop,
                        },
                        'op': original_filter.filters[1].inner_filter.operator,
                        'value': property_filters[1].value.to_repr(),
                    },
                },
            ],
        }

        output_filter = CompositeFilter.from_repr(data)

        assert output_filter == original_filter

    def test_composite_filter_to_repr(self, property_filters):
        composite_filter = CompositeFilter(
            operator=CompositeFilterOperator.AND,
            filters=[
                Filter(property_filters[0]),
                Filter(property_filters[1])
            ])
        query_filter = Filter(composite_filter)

        r = query_filter.to_repr()

        composite_filter_dict = r['compositeFilter']
        assert composite_filter_dict['op'] == 'AND'
        self._assert_is_correct_prop_dict_for_property_filter(
            composite_filter_dict['filters'][0]['propertyFilter'],
            property_filters[0])
        self._assert_is_correct_prop_dict_for_property_filter(
            composite_filter_dict['filters'][1]['propertyFilter'],
            property_filters[1])

    @staticmethod
    def test_filter_from_repr(composite_filter):
        original_filter = Filter(inner_filter=composite_filter)

        data = {
            'compositeFilter': original_filter.inner_filter.to_repr()
        }

        output_filter = Filter.from_repr(data)

        assert output_filter == original_filter

    @staticmethod
    def test_filter_from_repr_unexpected_filter_name():
        unexpected_filter_name = 'unexpectedFilterName'
        data = {
            unexpected_filter_name: 'DoesNotMatter'
        }

        with pytest.raises(ValueError) as ex_info:
            Filter.from_repr(data)

        assert unexpected_filter_name in ex_info.value.args[0]

    @staticmethod
    def test_filter_to_repr(composite_filter):
        test_filter = Filter(inner_filter=composite_filter)

        r = test_filter.to_repr()

        assert r['compositeFilter'] == test_filter.inner_filter.to_repr()

    @staticmethod
    def test_repr_returns_to_repr_as_string(query_filter):
        assert repr(query_filter) == str(query_filter.to_repr())

    @staticmethod
    @pytest.fixture(scope='session')
    def property_filters()                        :
        return [
            PropertyFilter(
                prop='prop1',
                operator=PropertyFilterOperator.LESS_THAN,
                value=Value('value1')
            ),
            PropertyFilter(
                prop='prop2',
                operator=PropertyFilterOperator.GREATER_THAN,
                value=Value(1234)
            )
        ]

    @staticmethod
    @pytest.fixture(scope='session')
    def composite_filter(property_filters)                   :
        return CompositeFilter(
            operator=CompositeFilterOperator.AND,
            filters=[
                Filter(property_filters[0]),
                Filter(property_filters[1])
            ])

    @staticmethod
    @pytest.fixture(scope='session')
    def query_filter(composite_filter)          :
        return Filter(inner_filter=composite_filter)

    @staticmethod
    @pytest.fixture(scope='session')
    def value()         :
        return Value('value')

    @staticmethod
    def _assert_is_correct_prop_dict_for_property_filter(
            prop_dict                , property_filter                ):
        assert prop_dict['property']['name'] == property_filter.prop
        assert prop_dict['op'] == property_filter.operator.value
        assert prop_dict['value'] == property_filter.value.to_repr()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from builtins import str
import warnings
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from future import standard_library
standard_library.install_aliases()
from builtins import object
import pytest
from gcloud.rest.datastore import LatLng


class TestLatLng(object):
    @staticmethod
    def test_from_repr(lat_lng):
        original_latlng = lat_lng
        data = {
            'latitude': original_latlng.lat,
            'longitude': original_latlng.lon,
        }

        output_order = LatLng.from_repr(data)

        assert output_order == original_latlng

    @staticmethod
    def test_to_repr():
        lat = 49.2827
        lon = 123.1207
        latlng = LatLng(lat, lon)

        r = latlng.to_repr()

        assert r['latitude'] == lat
        assert r['longitude'] == lon

    @staticmethod
    def test_repr_returns_to_repr_as_string(lat_lng):
        assert repr(lat_lng) == str(lat_lng.to_repr())

    @staticmethod
    @pytest.fixture(scope='session')
    def lat_lng()          :
        return LatLng(49.2827, 123.1207)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
import warnings
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from future import standard_library
standard_library.install_aliases()
import gcloud.rest.datastore.datastore as datastore  # pylint: disable=unused-import


def test_importable():
    assert True



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from builtins import str
import warnings
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from future import standard_library
standard_library.install_aliases()
from builtins import object
from typing import Any
from typing import Dict


# https://cloud.google.com/datastore/docs/reference/data/rest/Shared.Types/LatLng
class LatLng(object):
    def __init__(self, lat       , lon       )        :
        self.lat = lat
        self.lon = lon

    def __eq__(self, other     )        :
        if not isinstance(other, LatLng):
            return False

        return bool(
            self.lat == other.lat
            and self.lon == other.lon)

    def __repr__(self)       :
        return str(self.to_repr())

    @classmethod
    def from_repr(cls, data                )            :
        lat = data['latitude']
        lon = data['longitude']
        return cls(lat=lat, lon=lon)

    def to_repr(self)                  :
        return {
            'latitude': self.lat,
            'longitude': self.lon,
        }



import sys
from datetime import datetime

import pytest
from gcloud.rest.datastore import Key
from gcloud.rest.datastore import LatLng
from gcloud.rest.datastore import PathElement
from gcloud.rest.datastore import Value


class TestValue(object):
    @staticmethod
    @pytest.mark.parametrize('json_key,json_value', [
        ('booleanValue', True),
        ('doubleValue', 34.48),
        ('integerValue', 8483),
        ('stringValue', 'foobar'),
        ('booleanValue', False),
        ('doubleValue', 0.0),
        ('integerValue', 0),
        ('stringValue', ''),
    ])
    def test_from_repr(json_key, json_value):
        data = {
            'excludeFromIndexes': False,
            json_key: json_value
        }

        value = Value.from_repr(data)

        assert value.excludeFromIndexes is False
        assert value.value == json_value

    @staticmethod
    def test_from_repr_with_null_value():
        data = {
            'excludeFromIndexes': False,
            'nullValue': 'NULL_VALUE'
        }

        value = Value.from_repr(data)

        assert value.excludeFromIndexes is False
        assert value.value is None

    @staticmethod
    @pytest.mark.parametrize('v,expected', [
        ('1998-07-12T11:22:33.456789000Z',
         datetime(year=1998, month=7, day=12, hour=11,
                  minute=22, second=33, microsecond=456789)),
        ('1998-07-12T11:22:33.456789Z',
         datetime(year=1998, month=7, day=12, hour=11,
                  minute=22, second=33, microsecond=456789)),
        ('1998-07-12T11:22:33.456Z',
         datetime(year=1998, month=7, day=12, hour=11,
                  minute=22, second=33, microsecond=456000)),
        ('1998-07-12T11:22:33',
         datetime(year=1998, month=7, day=12, hour=11,
                  minute=22, second=33, microsecond=0)),
    ])
    def test_from_repr_with_datetime_value(v, expected):
        data = {
            'excludeFromIndexes': False,
            'timestampValue': v
        }

        value = Value.from_repr(data)
        assert value.value == expected

    @staticmethod
    @pytest.mark.skipif(sys.version_info[0] < 3,
                        reason='skipping because python2 has same '
                               'type for str and bytes')
    def test_from_repr_with_blob_value():
        data = {
            'excludedFromIndexed': False,
            'blobValue': 'Zm9vYmFy'
        }

        value = Value.from_repr(data)
        assert value.value == b'foobar'

    @staticmethod
    def test_from_repr_with_key_value(key):
        data = {
            'excludeFromIndexes': False,
            'keyValue': key.to_repr()
        }

        value = Value.from_repr(data)

        assert value.value == key

    @staticmethod
    def test_from_repr_with_geo_point_value(lat_lng):
        data = {
            'excludeFromIndexes': False,
            'geoPointValue': lat_lng.to_repr()
        }

        value = Value.from_repr(data)

        assert value.value == lat_lng

    @staticmethod
    def test_from_repr_could_not_find_supported_value_key():
        data = {
            'excludeFromIndexes': False,
        }

        with pytest.raises(NotImplementedError) as ex_info:
            Value.from_repr(data)

        assert 'excludeFromIndexes' in ex_info.value.args[0]

    @staticmethod
    @pytest.mark.parametrize('v,expected_json_key', [
        (True, 'booleanValue'),
        (34.48, 'doubleValue'),
        (8483, 'integerValue'),
        ('foobar', 'stringValue'),
        (False, 'booleanValue'),
        (0.0, 'doubleValue'),
        (0, 'integerValue'),
        ('', 'stringValue'),
    ])
    def test_to_repr(v, expected_json_key):
        value = Value(v)

        r = value.to_repr()

        assert len(r) == 2  # Value + excludeFromIndexes
        assert r['excludeFromIndexes'] is False
        assert r[str(expected_json_key)] == v

    @staticmethod
    def test_to_repr_with_null_value():
        value = Value(None)

        r = value.to_repr()

        assert r['nullValue'] == 'NULL_VALUE'

    @staticmethod
    def test_to_repr_with_datetime_value():
        dt = datetime(year=2018, month=7, day=15, hour=11, minute=22,
                      second=33, microsecond=456789)
        value = Value(dt)

        r = value.to_repr()

        assert r['timestampValue'] == '2018-07-15T11:22:33.456789000Z'

    @pytest.mark.skipif(sys.version_info[0] < 3,
                        reason='skipping because python2 has same '
                               'type for str and bytes')
    @staticmethod
    def test_to_repr_with_blob_value():
        value = Value(b'foobar')

        r = value.to_repr()
        assert r['blobValue'] == 'Zm9vYmFy'

    @staticmethod
    def test_to_repr_with_key_value(key):
        value = Value(key)

        r = value.to_repr()

        assert r['keyValue'] == key.to_repr()

    @staticmethod
    def test_to_repr_with_custom_key_value():
        class CustomKey(object):
            @staticmethod
            def to_repr():
                return {'some': 'JSON'}

        class CustomValue(Value):
            key_kind = CustomKey

        key = CustomKey()
        value = CustomValue(key)

        r = value.to_repr()

        assert r['keyValue'] == key.to_repr()

    @staticmethod
    def test_to_repr_with_geo_point_value(lat_lng):
        value = Value(lat_lng)

        r = value.to_repr()

        assert r['geoPointValue'] == lat_lng.to_repr()

    @staticmethod
    def test_to_repr_exclude_from_indexes():
        value = Value(123, exclude_from_indexes=True)

        r = value.to_repr()

        assert r['excludeFromIndexes']

    @staticmethod
    def test_to_repr_non_supported_type():
        class NonSupportedType(object):
            pass

        value = Value(NonSupportedType())

        with pytest.raises(Exception) as ex_info:
            value.to_repr()

        assert NonSupportedType.__name__ in ex_info.value.args[0]

    @staticmethod
    def test_repr_returns_to_repr_as_string(value):
        assert repr(value) == str(value.to_repr())

    @staticmethod
    @pytest.fixture(scope='session')
    def key()       :
        path = PathElement(kind='my-kind', name='path-name')
        key = Key(project='my-project', path=[path], namespace='my-namespace')
        return key

    @staticmethod
    @pytest.fixture(scope='session')
    def lat_lng()          :
        return LatLng(49.2827, 123.1207)

    @staticmethod
    @pytest.fixture(scope='session')
    def value()         :
        return Value(value='foobar', exclude_from_indexes=False)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from builtins import str
import warnings
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from future import standard_library
standard_library.install_aliases()
from builtins import object
from typing import Any
from typing import Dict


# https://cloud.google.com/datastore/docs/reference/data/rest/v1/projects/runQuery#Projection
class Projection(object):
    def __init__(self, prop     )        :
        self.prop = prop

    def __eq__(self, other     )        :
        if not isinstance(other, Projection):
            return False

        return bool(self.prop == other.prop)

    def __repr__(self)       :
        return str(self.to_repr())

    @classmethod
    def from_repr(cls, data                )                :
        return cls(prop=data['property']['name'])

    def to_repr(self)                  :
        return {
            'property': {'name': self.prop},
        }
